#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 12 Oct 2023 10:34
# @Author: Yao LI
# @File: spagrn/prc.py


import os
from typing import Union
import glob
import json
import anndata
import scanpy as sc
import pandas as pd
from scipy import stats
import matplotlib.pyplot as plt
import seaborn as sns
from itertools import product
from sklearn.metrics import precision_recall_curve, roc_curve, auc

import matplotlib as mpl

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['svg.fonttype'] = 'none'
'''
python ../../../../spagrn_debug/au.py
'''


class AUPRC:
    def __init__(self, tfs=None, data=None, adj=None, name_df=None):
        self._adata = data  # only necessary when using spearman cc
        self._tfs = tfs

        self._ground_truth = None  # all genes
        self._prediction = None  # all genes
        self._baseline = None
        self._auprc_ratio = None
        self._auprc = None
        self._auroc = None

        self._adj = adj
        self._regulons = None

        # column names in Ground Truth file
        self._value_col = 'regulator.effect'
        self._tf_col = 'regulator.gene'
        self._target_col = 'regulated.gene'

        self._name_df = name_df  # ID-gene name chart
        self._true_df = None  # ground truth exclude noises
        self._pred_df = None  # prediction exclude genes that has been filtered

        self.prec = None
        self.recall = None
        self.thresholds = None

    @property
    def adata(self):
        return self._adata

    @adata.setter
    def adata(self, value):
        self._adata = value

    @property
    def tfs(self):
        return self._tfs

    @property
    def name_df(self):
        return self._name_df

    @name_df.setter
    def name_df(self, value):
        self._name_df = value

    @property
    def ground_truth(self):
        return self._ground_truth

    @ground_truth.setter
    def ground_truth(self, value):
        self._ground_truth = value

    @property
    def true_df(self):
        return self._true_df

    @true_df.setter
    def true_df(self, value):
        self._true_df = value

    @property
    def pred_df(self):
        return self._pred_df

    @pred_df.setter
    def pred_df(self, value):
        self._pred_df = value

    @property
    def baseline(self):
        return self._baseline

    @baseline.setter
    def baseline(self, value):
        self._baseline = value

    @property
    def auprc_ratio(self):
        return self._auprc_ratio

    @auprc_ratio.setter
    def auprc_ratio(self, value):
        self._auprc_ratio = value

    @property
    def auroc(self):
        return self._auroc

    @auroc.setter
    def auroc(self, value):
        self._auroc = value

    @property
    def auprc(self):
        return self._auprc

    @auprc.setter
    def auprc(self, value):
        self._auprc = value

    @property
    def prediction(self):
        return self._prediction

    @prediction.setter
    def prediction(self, value):
        self._prediction = value

    @property
    def value_col(self):
        return self._value_col

    @property
    def tf_col(self):
        return self._tf_col

    @property
    def target_col(self):
        return self._target_col

    @target_col.setter
    def target_col(self, value):
        self._target_col = value

    @property
    def adj(self):
        return self._adj

    @adj.setter
    def adj(self, value):
        self._adj = value

    @property
    def regulons(self):
        return self._regulons

    @regulons.setter
    def regulons(self, value):
        self._regulons = value

    def get_true_df(self, ground_truth_files):
        """

        :param ground_truth_files:
        :return:
        """
        fl = glob.glob(ground_truth_files)
        self.true_df = pd.concat([pd.read_csv(i) for i in fl]).astype(str)
        return self.true_df

    # def make_ground_truth(self, ground_truth_files, real_tfs=None, false_tfs=None):
    #     """
    #
    #     :param ground_truth_files:
    #     :param real_tfs:
    #     :param false_tfs:
    #     :return:
    #     """
    #     # names = pd.read_csv(naming_fn)
    #     fl = glob.glob(ground_truth_files)
    #     df_true = pd.concat([pd.read_csv(i) for i in fl]).astype(str)
    #
    #     # adata = sc.read_h5ad(adata_fn)
    #     # if self.adata is None:
    #     #     self.adata = adata
    #     all_genes = self.adata.var_names
    #     ground_truth = pd.DataFrame(product(self.tfs, all_genes), columns=['regulator.gene', 'regulated.gene']).astype(
    #         str)
    #     # ! make sure gene names are using the same nomenclature
    #     ground_truth['regulated.gene'] = ground_truth['regulated.gene'].replace(list(self.name_df['name']),
    #                                                                             list(self.name_df['id']))
    #     ground_truth['regulator.effect'] = [0] * ground_truth.shape[0]
    #     ground_truth = pd.concat([ground_truth, df_true])
    #     ground_truth = ground_truth.drop_duplicates(['regulator.gene', 'regulated.gene'], keep='last')
    #
    #     # if false TF exists
    #     if real_tfs and false_tfs:
    #         t_ground_truth = ground_truth[ground_truth['regulator.gene'].isin(real_tfs)]
    #         f_ground_truth = ground_truth[ground_truth['regulator.gene'].isin(false_tfs)]
    #         f_ground_truth['regulator.effect'] = [0.0] * f_ground_truth.shape[0]
    #         ground_truth = pd.concat([t_ground_truth, f_ground_truth])
    #
    #     ground_truth[['regulator.gene', 'regulated.gene']] = ground_truth[['regulator.gene', 'regulated.gene']].replace(
    #         list(self.name_df['id']), list(self.name_df['name']))
    #     ground_truth['regulator.effect'] = ground_truth['regulator.effect'].astype('float64')
    #     # convert y_true into a binary matrix
    #     ground_truth.loc[ground_truth['regulator.effect'] > 0, 'regulator.effect'] = 1
    #     # order of genes need to be consistent between ground_truth and prediction
    #     ground_truth = ground_truth.sort_values(['regulator.gene', 'regulated.gene'], ascending=[True, True])
    #     self.ground_truth = ground_truth
    #     return ground_truth

    def get_baseline(self):
        self.baseline = 1 - self.ground_truth[self.ground_truth[self.value_col] == 0].shape[0] / \
                        self.ground_truth.shape[0]
        print(f'Baseline is {self.baseline} (num of 1/num of total)')
        return self.baseline

    def get_pred_df(self, y_true_label=None, y_true_tf_col=None, y_true_target_col=None):
        """

        :param y_true_label:
        :return:
        """
        if self.adj is None:
            self.adj = self.adata.uns['adj']
        if y_true_label is None:
            y_true_label = self.value_col
        if y_true_tf_col is None:
            y_true_tf_col = self.tf_col
        if y_true_target_col is None:
            y_true_target_col = self.target_col

        # 2. input prediction value
        self.regulons = self.adata.uns['regulon_dict']
        mylist = [(key, x) for key, val in self.regulons.items() for x in val]
        df_pred = pd.DataFrame(mylist, columns=['Name', 'Values'])
        # TODO: if has (+)
        df_pred['Name'] = df_pred['Name'].str.strip('(+)')
        df_pred['prediction'] = [1] * df_pred.shape[0]

        # 1. get importance (pred_label) values
        df_pred = self.adj.merge(df_pred, left_on=['TF', 'target'], right_on=['Name', 'Values'], how='left')
        df_pred['prediction'] = df_pred['prediction'].fillna(0)

        # 3. introduce ground truth classification label
        df_pred = df_pred.merge(self.ground_truth, left_on=['TF', 'target'],
                                right_on=[y_true_tf_col, y_true_target_col],
                                how='left')

        df_pred = df_pred[['TF', 'target', 'importance', 'prediction', y_true_label]]
        df_pred.columns = [y_true_tf_col, y_true_target_col, 'importance', 'prediction', 'ground truth']
        # df_pred.to_csv('df_pred.csv', index=False)
        self.pred_df = df_pred
        return self.pred_df

    def get_pred_df_grnboost(self, y_true_label=None, y_true_tf_col=None, y_true_target_col=None):
        """

        :param y_true_label:
        :return:
        """
        if self.adj is None:
            self.adj = self.adata.uns['adj']
        if y_true_label is None:
            y_true_label = self.value_col
        if y_true_tf_col is None:
            y_true_tf_col = self.tf_col
        if y_true_target_col is None:
            y_true_target_col = self.target_col

        # 1. input prediction value
        df_pred = self.adj.copy()
        df_pred['prediction'] = [1] * df_pred.shape[0]

        # 3. introduce ground truth classification label
        df_pred = df_pred.merge(self.ground_truth, left_on=['TF', 'target'],
                                right_on=[y_true_tf_col, y_true_target_col],
                                how='left')

        df_pred = df_pred[['TF', 'target', 'importance', 'prediction', y_true_label]]
        df_pred.columns = [y_true_tf_col, y_true_target_col, 'importance', 'prediction', 'ground truth']
        # df_pred.to_csv('df_pred.csv', index=False)
        self.pred_df = df_pred
        return self.pred_df

    # alternative to get_pred_df
    # calculate spearman values
    def get_pred_df_spearman(self, data: Union[pd.DataFrame, anndata.AnnData] = None, y_true_label=None,
                             y_true_tf_col=None, y_true_target_col=None):
        """

        :param data:
        :param y_true_label:
        :param y_true_tf_col:
        :param y_true_target_col:
        :return:
        """
        if y_true_label is None:
            y_true_label = self.value_col
        if y_true_tf_col is None:
            y_true_tf_col = self.tf_col
        if y_true_target_col is None:
            y_true_target_col = self.target_col
        # 1. calculate spearman cc
        # adata = sc.read_h5ad(self.adata_fn)
        # if self.adata is None:
        #     self.adata = adata
        # df = adata.to_df()
        if data is None:
            raise ValueError('Must provide expression matrix when pred label is set to spearman')
        if isinstance(data, anndata.AnnData):
            data = data.to_df()
        if self.adj is None:
            self.adj = self.adata.uns['adj']
        s = []
        for i in self.adj.index:
            res = stats.spearmanr(data[self.adj.loc[i].TF], data[self.adj.loc[i].target])
            s.append(res.correlation)
        self.adj['spearman'] = s
        adj = self.adj.sort_values(['importance', 'spearman'], ascending=False)

        # 2. input prediction value
        regs = self.adata.uns['regulon_dict']
        mylist = [(key, x) for key, val in regs.items() for x in val]
        df_pred = pd.DataFrame(mylist, columns=['Name', 'Values'])
        df_pred['Name'] = df_pred['Name'].str.strip('(+)')
        df_pred['prediction'] = [1] * df_pred.shape[0]

        # 1. merge spearman df and prediction df
        df_pred = adj.merge(df_pred, left_on=['TF', 'target'], right_on=['Name', 'Values'], how='left')
        df_pred['prediction'].fillna(0)
        df_pred['prediction'] = df_pred['prediction'].fillna(0)

        # 3. introduce ground truth classification label
        df_pred = df_pred.merge(self.ground_truth, left_on=['TF', 'target'],
                                right_on=[y_true_tf_col, y_true_target_col],
                                how='left')
        df_pred = df_pred[['TF', 'target', 'importance', 'spearman', 'prediction', y_true_label]]
        # sort by spearman value
        tt1 = df_pred[df_pred.prediction > 0]
        tt0 = df_pred[df_pred.prediction == 0]
        tt1 = tt1.sort_values(['spearman'], ascending=False)
        tt0 = tt0.sort_values(['spearman'], ascending=False)
        # make sure 0 labels (negatives) spearman value is smaller than 1 labels
        tt0['spearman'] = tt0['spearman'] - 1
        df_prediction = pd.concat([tt1, tt0])
        df_prediction.columns = [y_true_tf_col, y_true_target_col, 'importance', 'spearman', 'prediction',
                                 'ground truth']
        # df_prediction.to_csv('df_pred.csv', index=False)
        self.pred_df = df_prediction
        # self.adata.uns['prediction'] = df_prediction
        return df_prediction

    def get_prediction_df(self, pred_label='spearman', y_true_tf_col=None, y_true_target_col=None):
        """
        get prediction for all genes (including genes had been filtered out by SpaGRN),
        so Ground Truth and Prediction have the same dimension (aka len(all_genes))
        value_col: column of value to pass in AUPRC calculation e.g. importance, spearman coefficient ...
        :param pred_label:
        :param y_true_tf_col:
        :param y_true_target_col:
        :return:
        """
        if y_true_tf_col is None:
            y_true_tf_col = self.tf_col
        if y_true_target_col is None:
            y_true_target_col = self.target_col

        pred_index = pd.merge(
            self.pred_df[[y_true_tf_col, y_true_target_col, pred_label, 'prediction', 'ground truth']],
            self.ground_truth[[y_true_tf_col, y_true_target_col]], on=[y_true_tf_col, y_true_target_col],
            how='outer')
        assert pred_index.shape[0] == self.ground_truth.shape[0]
        pred = pred_index.sort_values([y_true_tf_col, y_true_target_col], ascending=[True, True])
        if pred_label == 'spearman':
            pred = pred.fillna(int(pred[pred_label].min()) - 2)
        else:
            pred = pred.fillna(0)

        self.prediction = pred
        return self.prediction

    def get_auprc(self, y_true_label='regulator.effect', pred_label='spearman'):
        """

        :param y_true_label:
        :param pred_label:
        :return:
        """
        self.prec, self.recall, self.thresholds = precision_recall_curve(y_true=self.ground_truth[y_true_label],
                                                                         probas_pred=self.prediction[pred_label],
                                                                         pos_label=1)
        new_auc = auc(self.recall, self.prec)
        self.auprc = new_auc

    def get_ratio(self):
        """
        AUPRC ratio
        :return:
        """
        if self.baseline:  # walrus operator cannot be used on instance attributes?
            self.auprc_ratio = self.auprc / self.baseline
        else:
            self.get_baseline()
            self.auprc_ratio = self.auprc / self.baseline
        if self.adata is not None and isinstance(self.adata, anndata.AnnData):
            self.adata.uns['auprc_ratio'] = self.auprc_ratio
        print(f'AUPRC ratio is {self.auprc_ratio}.')

    def plot_prec_recall(self, fn='Precision-Recall.png'):
        if self.recall is None or self.prec is None:
            raise ValueError('Calculate prc first plotting. See method get_auprc')
        plt.fill_between(self.recall, self.prec)
        plt.ylabel("Precision")
        plt.xlabel("Recall")
        plt.title("Train Precision-Recall curve")
        plt.savefig(fn)
        plt.close()

    def get_auroc(self, y_true_label='regulator.effect', pred_label='spearman'):
        """

        :param y_true_label:
        :param pred_label:
        :return:
        """
        fpr, tpr, thresholds2 = roc_curve(y_true=self.ground_truth[y_true_label],
                                          y_score=self.prediction[pred_label],
                                          pos_label=1)
        auroc = auc(fpr, tpr)
        self.auroc = auroc
        return auroc
        # plt.fill_between(fpr, tpr)
        # plt.ylabel("true positive")
        # plt.xlabel("false positive")
        # plt.title("AUROC")
        # plt.savefig('aucroc.png')
        # plt.close()

    def prc(self,
            pred_label, ground_truth: pd.DataFrame,
            y_true_label='regulator.effect',
            fn='adata.h5ad',
            fig_fn='Precision-Recall.png'):
        """
        Main logic method. SpaGRN.AUPRC pipeline
        1. generate ground truth
        2.1. load in prediction output by SpaGRN
        2.2. fill in the blank
        3. Calculate AUPRC and plot result
        :param pred_label:
        :param ground_truth_files:
        :param y_true_label:
        :param fn:
        :return:
        """
        # 1.
        # self.make_ground_truth(ground_truth_files, real_tfs=['2', '232', '408', '805', '1006'],
        #                        false_tfs=['1140', '1141', '1142', '1143', '1144'])
        # self.get_baseline()
        self.ground_truth = ground_truth

        # 2.
        if pred_label == 'spearman':
            self.get_pred_df_spearman(data=self.adata)
        else:
            # self.adj = adj_df
            self.get_pred_df_grnboost()
            # self.get_pred_df()
        self.get_prediction_df(pred_label=pred_label)

        # 3.
        self.get_auprc(pred_label=pred_label, y_true_label=y_true_label)
        self.get_ratio()
        # self.plot_prec_recall(fn=fig_fn)

        # 4. save results
        # self.adata.write_h5ad(fn)

    def roc(self,
            pred_label,
            ground_truth: pd.DataFrame,
            adj_df=None,
            y_true_label='regulator.effect',
            fn='adata.h5ad',
            fig_fn='Precision-Recall.png'):
        """
        Main logic method. SpaGRN.AUPRC pipeline
        1. generate ground truth
        2.1. load in prediction output by SpaGRN
        2.2. fill in the blank
        3. Calculate AUPRC and plot result
        :param pred_label:
        :param ground_truth_files:
        :param y_true_label:
        :param fn:
        :return:
        """
        # 1.
        # self.make_ground_truth(ground_truth_files, real_tfs=['2', '232', '408', '805', '1006'],
        #                        false_tfs=['1140', '1141', '1142', '1143', '1144'])
        # self.get_baseline()
        self.ground_truth = ground_truth

        # 2.
        if pred_label == 'spearman':
            self.get_pred_df_spearman(data=self.adata)
        else:
            # self.adj = adj_df
            self.get_pred_df_grnboost()
            # self.get_pred_df()
        self.get_prediction_df(pred_label=pred_label)

        # 4.
        self.get_auroc(pred_label=pred_label, y_true_label=y_true_label)

        # 5. save results
        # self.adata.write_h5ad(fn)


def cal_auprc(tfs, name_df, ground_truth, pred_label='spearman', adata=None, adj_fn=None, fn='adata.h5ad'):
    """

    :param adata:
    :param tfs:
    :param name_df:
    :param adj_fn:
    :param reg_fn:
    :param pred_label:
    :param ground_truth_files:
    :return:
    """
    a = AUPRC(data=adata,
              adj=adj_fn,
              tfs=tfs,
              name_df=name_df)
    a.prc(pred_label=pred_label, ground_truth=ground_truth, fn=fn)
    return a.auprc_ratio


def cal_auroc(tfs, name_df, ground_truth, pred_label='spearman', adata=None, adj_fn=None, fn='adata.h5ad'):
    """

    :param adata:
    :param tfs:
    :param name_df:
    :param adj_fn:
    :param reg_fn:
    :param pred_label:
    :param ground_truth_files:
    :return:
    """
    a = AUPRC(data=adata,
              adj=adj_fn,
              tfs=tfs,
              name_df=name_df)
    a.roc(pred_label=pred_label, ground_truth=ground_truth, fn=fn)
    return a.auroc


def make_ground_truth(ground_truth_files, name_df: pd.DataFrame, all_genes: list, real_tfs=None, false_tfs=None):
    """

    :param ground_truth_files:
    :param name_df:
    :param all_genes:
    :param real_tfs:
    :param false_tfs:
    :return:
    """
    # names = pd.read_csv(naming_fn)
    fl = glob.glob(ground_truth_files)
    df_true = pd.concat([pd.read_csv(i) for i in fl]).astype(str)

    # adata = sc.read_h5ad(adata_fn)
    # if self.adata is None:
    #     self.adata = adata
    # all_genes = adata.var_names
    ground_truth = pd.DataFrame(product(tfs, all_genes), columns=['regulator.gene', 'regulated.gene']).astype(
        str)
    # ! make sure gene names are using the same nomenclature
    ground_truth['regulated.gene'] = ground_truth['regulated.gene'].replace(list(name_df['name']),
                                                                            list(name_df['id']))
    ground_truth['regulator.effect'] = [0] * ground_truth.shape[0]
    ground_truth = pd.concat([ground_truth, df_true])
    ground_truth = ground_truth.drop_duplicates(['regulator.gene', 'regulated.gene'], keep='last')

    # if false TF exists
    if real_tfs and false_tfs:
        t_ground_truth = ground_truth[ground_truth['regulator.gene'].isin(real_tfs)]
        f_ground_truth = ground_truth[ground_truth['regulator.gene'].isin(false_tfs)]
        f_ground_truth['regulator.effect'] = [0.0] * f_ground_truth.shape[0]
        ground_truth = pd.concat([t_ground_truth, f_ground_truth])

    ground_truth[['regulator.gene', 'regulated.gene']] = ground_truth[['regulator.gene', 'regulated.gene']].replace(
        list(name_df['id']), list(name_df['name']))
    ground_truth['regulator.effect'] = ground_truth['regulator.effect'].astype('float64')
    # convert y_true into a binary matrix
    ground_truth.loc[ground_truth['regulator.effect'] > 0, 'regulator.effect'] = 1
    # order of genes need to be consistent between ground_truth and prediction
    ground_truth = ground_truth.sort_values(['regulator.gene', 'regulated.gene'], ascending=[True, True])
    return ground_truth


def calculate_multi_samples(methods, data_fn_base, tfs, names, ver7_ground_truth):
    data_nums = list(range(2, 12))
    ratios = {}
    aurocs = {}
    for method in methods:
        ratios[method] = []
        aurocs[method] = []
        for num in data_nums:
            data_folder = os.path.join(data_fn_base, f'data{num}')
            if method == 'hotspot':
                data_fn = os.path.join(data_folder, f'{method}/{method}_spagrn.h5ad')
                adata = sc.read_h5ad(data_fn)
                pred_label = 'spearman'
                ratio = cal_auprc(tfs, names, adata=adata, pred_label=pred_label, ground_truth=ver7_ground_truth)
            elif method == 'HOTSPOT':
                data_fn = os.path.join(data_folder, f'{method}/hotspot.h5ad')
                adata = sc.read_h5ad(data_fn)
                pred_label = 'importance'
                ratio = cal_auprc(tfs, names, adata=adata, pred_label=pred_label, ground_truth=ver7_ground_truth)
            elif method == 'genie3':
                data_fn = os.path.join(data_folder, f'{method}/genie3.adj.csv')
                df = pd.read_csv(data_fn)
                ratio = cal_auprc(tfs, names, pred_label=pred_label, ground_truth=ver7_ground_truth, adj_fn=df)
                pred_label = 'importance'
            elif method == 'grnboost':
                data_fn = os.path.join(data_folder, f'{method}/grnboost_adj.csv')
                df = pd.read_csv(data_fn)
                ratio = cal_auprc(tfs, names, pred_label=pred_label, ground_truth=ver7_ground_truth, adj_fn=df)
                pred_label = 'importance'

            # adata = sc.read_h5ad(data_fn)
            ratios[method].append(ratio)
            auroc = cal_auroc(adata, tfs, names, pred_label=pred_label, ground_truth=ver7_ground_truth)
            aurocs[method].append(auroc)

    with open('ratios.json_sub', 'w') as f:
        json.dump(ratios, f, sort_keys=True, indent=4)
    with open('aurocs.json_sub', 'w') as f:
        json.dump(aurocs, f, sort_keys=True, indent=4)
    return ratios, aurocs


def ratios_boxplot(ratios, methods, x_labels=None, fn='auprc_ratio_boxplot.pdf'):
    if x_labels is None:
        x_labels = ['SpaGRN', 'GRNBoost2', 'GENIE3', 'HOTSPOT']
    df = pd.DataFrame.from_records(ratios).astype('float64')
    df = df[methods]
    ax = sns.boxplot(data=df)
    plt.title('AUPRC ratios')
    plt.ylabel('ratio')
    ax.set_xticklabels(x_labels)
    plt.tight_layout()
    plt.savefig(fn, format='pdf')
    plt.close()


def auroc_boxplot(aurocs, methods, x_labels=None, fn='auroc_ratio_boxplot.pdf'):
    if x_labels is None:
        x_labels = ['SpaGRN', 'GRNBoost2', 'GENIE3', 'HOTSPOT']
    df = pd.DataFrame.from_records(aurocs).astype('float64')
    df = df[methods]
    ax = sns.boxplot(data=df)
    plt.title('AUROC')
    plt.ylabel('auroc')
    ax.set_xticklabels(x_labels)
    plt.tight_layout()
    plt.savefig(fn, format='pdf')
    plt.close()


if __name__ == '__main__':
    # names = pd.read_csv('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver7/name_df.csv')
    '''
    print(names)
    id,name
    2,Adf1
    232,Aef1
    408,grh
    '''
    # tfs = [2, 232, 408, 805, 1006, 1140, 1141, 1142, 1143, 1144]
    # ver7_ground_truth = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver7/GRN_params_*.csv'
    # data_fn_base = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8'

    # for one sample
    # 2023-10-24: test h5ad adj & regulons: success
    # adata = sc.read_h5ad('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/data11/HOTSPOT.h5ad')
    # adata = sc.read_h5ad('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/data11/hetero_11.h5ad')
    # test run for GENIE3
    # adj_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/data11/genie3.adj.csv'
    # adj_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/grnboost_adj.csv'
    # reg_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/data11/grnboost/grnboost_regulons.json'
    # ratio = cal_auprc(adata, tfs, names, adj_fn=adj_fn, reg_fn=reg_fn, pred_label='importance', ground_truth_files=ver7_ground_truth)
    # 2023-10-24
    # ratio = cal_auprc(adata, tfs, names, pred_label='importance', ground_truth_files=ver7_ground_truth)
    # print(ratio)

    # if os.path.isfile('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/ratios.json') \
    #         and os.path.isfile('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/aurocs.json'):
    #     rs = json.load(open('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/ratios.json'))
    #     aus = json.load(open('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/aurocs.json'))
    # else:
    # 2023-10-26: final version
    # methods = ['hotspot', 'grnboost', 'genie3', 'HOTSPOT']
    # rs, aus = calculate_multi_samples(methods)
    # ratios_boxplot(rs, methods)
    # auroc_boxplot(aus, methods)

    # 2023-10-25, 27
    # 补run HOTSPOT
    # ratios = json.load(open('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/ratios.json'))
    # aurocs = json.load(open('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/aurocs.json'))
    # data_nums = list(range(2, 12))
    # methods = ['HOTSPOT']
    # for method in methods:
    #     ratios[method] = []
    #     aurocs[method] = []
    #     for num in data_nums:
    #         data_folder = os.path.join(data_fn_base, f'data{num}')
    #         data_fn = os.path.join(data_folder, f'{method}/hotspot.h5ad')
    #         if method == 'hotspot':
    #             pred_label = 'spearman'
    #         else:
    #             pred_label = 'importance'
    #
    #         adata = sc.read_h5ad(data_fn)
    #         ratio = cal_auprc(adata, tfs, names, pred_label=pred_label, ground_truth_files=ver7_ground_truth)
    #         ratios[method].append(ratio)
    #         auroc = cal_auroc(adata, tfs, names, pred_label=pred_label, ground_truth_files=ver7_ground_truth)
    #         aurocs[method].append(auroc)
    # with open('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/ratios.json', 'w') as f:
    #     json.dump(ratios, f, sort_keys=True, indent=4)
    # with open('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/aurocs.json', 'w') as f:
    #     json.dump(aurocs, f, sort_keys=True, indent=4)
    #
    # ratios_boxplot(ratios, methods=['hotspot', 'grnboost', 'genie3', 'HOTSPOT'])
    # auroc_boxplot(aurocs, methods=['hotspot', 'grnboost', 'genie3', 'HOTSPOT'])

    # 2023-11-13: run subdata results
    names = pd.read_csv('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver7/name_df.csv')
    # ver7_ground_truth = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver7/GRN_params_*.csv'
    adata = sc.read_h5ad('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/data11/hetero_11.h5ad')
    tfs = [2, 232, 408, 805, 1006, 1140, 1141, 1142, 1143, 1144]
    # a = AUPRC(data=adata,
    #           tfs=tfs,
    #           name_df=names)
    # gt = a.make_ground_truth(ver7_ground_truth, real_tfs=['2', '232', '408', '805', '1006'], false_tfs=['1140', '1141', '1142', '1143', '1144'])
    # print(gt)
    # gt.to_csv('ground_truth_ver7.csv', index=False)
    ver7_ground_truth_fn = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/ground_truth_ver7.csv'
    ver7_ground_truth = pd.read_csv(ver7_ground_truth_fn)

    data_fn_base200 = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/subdata_200/'
    data_fn_base100 = '/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/subdata_100/'

    # ratios = {}
    # aurocs = {}
    # data_nums = list(range(2, 12))
    # methods = ['hotspot', 'HOTSPOT']
    # for method in methods:
    #     ratios[method] = []
    #     aurocs[method] = []
    #     for num in data_nums:
    #         data_folder = os.path.join(data_fn_base100, f'data{num}')
    #         data_fn = os.path.join(data_folder, f'{method}.h5ad')
    #         if method == 'hotspot':
    #             pred_label = 'spearman'
    #         else:
    #             pred_label = 'importance'
    #         adata = sc.read_h5ad(data_fn)
    #         ratio = cal_auprc(adata, tfs, names, pred_label=pred_label, ground_truth=ver7_ground_truth)
    #         ratios[method].append(ratio)
    #         auroc = cal_auroc(adata, tfs, names, pred_label=pred_label, ground_truth=ver7_ground_truth)
    #         aurocs[method].append(auroc)
    # with open(f'{data_fn_base100}/ratios.json', 'w') as f:
    #     json.dump(ratios, f, sort_keys=True, indent=4)
    # with open(f'{data_fn_base100}/aurocs.json', 'w') as f:
    #     json.dump(aurocs, f, sort_keys=True, indent=4)

    # ratios = json.load(open(f'{data_fn_base200}/ratios.json'))
    # aurocs = json.load(open(f'{data_fn_base200}/aurocs.json'))
    # data_nums = list(range(2, 12))
    # methods = ['grnboost', 'genie3']
    # for method in methods:
    #     ratios[method] = []
    #     aurocs[method] = []
    #     for num in data_nums:
    #         data_folder = os.path.join(data_fn_base200, f'data{num}')
    #         data_fn = os.path.join(data_folder, f'{method}.csv')
    #         if method == 'hotspot':
    #             pred_label = 'spearman'
    #         else:
    #             pred_label = 'importance'
    #         df = pd.read_csv(data_fn)
    #         ratio = cal_auprc(tfs, names, pred_label=pred_label, ground_truth=ver7_ground_truth, adj_fn=df)
    #         ratios[method].append(ratio)
    #         auroc = cal_auroc(tfs, names, pred_label=pred_label, ground_truth=ver7_ground_truth, adj_fn=df)
    #         aurocs[method].append(auroc)
    # with open(f'{data_fn_base200}/ratios.json', 'w') as f:
    #     json.dump(ratios, f, sort_keys=True, indent=4)
    # with open(f'{data_fn_base200}/aurocs.json', 'w') as f:
    #     json.dump(aurocs, f, sort_keys=True, indent=4)
    def list_insert(my_list,n_blanks=2):
        insert_positions = range(n_blanks, len(my_list) + n_blanks * (len(my_list) // 10 + 1), 10)
        new_list = []
        for i, val in enumerate(my_list):
            new_list.append(val)
            if i + 1 in insert_positions:
                new_list.extend([None] * n_blanks)
        return new_list


    ratios = json.load(open('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/ratios.json'))
    aurocs = json.load(open('/dellfsqd2/ST_OCEAN/USER/liyao1/07.spatialGRN/exp/07.simulation/ver8/aurocs.json'))
    ratios200 = json.load(open(f'{data_fn_base200}/ratios.json'))
    aurocs200 = json.load(open(f'{data_fn_base200}/aurocs.json'))
    ratios100 = json.load(open(f'{data_fn_base100}/ratios.json'))
    aurocs100 = json.load(open(f'{data_fn_base100}/aurocs.json'))

    def plot_sub(ratios,ratios200,ratios100,fn='boxplot_subset.pdf'):
        x_labels = ['SpaGRN(1500)', 'SpaGRN(1000)', 'SpaGRN(500)',
                    'GRNBoost2(1500)', 'GRNBoost2(1000)', 'GRNBoost2(500)',
                    'GENIE3(1500)', 'GENIE3(1000)', 'GENIE3(500)',
                    'HOTSPOT(1500)', 'HOTSPOT(1000)', 'HOTSPOT(500)']
        df = pd.DataFrame.from_records(ratios).astype('float64')
        df = df[['hotspot', 'grnboost', 'genie3', 'HOTSPOT']]  # set order
        df200 = pd.DataFrame.from_records(ratios200).astype('float64')
        df200 = df200[['hotspot', 'grnboost', 'genie3', 'HOTSPOT']]  # set order
        df200.columns = ['hotspot200', 'grnboost200', 'genie3200', 'HOTSPOT200']
        df100 = pd.DataFrame.from_records(ratios100).astype('float64')
        df100 = df100[['hotspot', 'grnboost', 'genie3', 'HOTSPOT']]  # set order
        df100.columns = ['hotspot100', 'grnboost100', 'genie3100', 'HOTSPOT100']
        df = pd.concat([df, df200, df100])
        df = df[['hotspot', 'hotspot200', 'hotspot100',
                 'grnboost', 'grnboost200', 'grnboost100',
                 'genie3', 'genie3200', 'genie3100',
                 'HOTSPOT', 'HOTSPOT200', 'HOTSPOT100']]

        # Reshape the dataframe using melt()
        df_melted = df.melt(var_name='Group', value_name='Value')
        # cate = ['A'] * 90 + ['B'] * 90 + ['C'] * 90 + ['D'] * 90
        # df_melted['category'] = cate
        #
        colors = ['green']*3 + ['orange']*3+['blue']*3+['red']*3
        # # colors = ['green','orange','blue','red']
        sns.set_palette(colors)
        # # ax = sns.boxplot(data=df)
        ax = sns.boxplot(x='Group', y='Value', data=df_melted)

        plt.title('AUPRC ratios')
        plt.ylabel('ratio')
        ax.set_xticklabels(x_labels)
        plt.xticks(rotation=45)
        plt.tight_layout()
        plt.savefig(fn, format='pdf')
        plt.close()


    plot_sub(aurocs,aurocs200,aurocs100)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 26 Oct 2023 16:53
# @Author: Yao LI
# @File: spagrn/simulator.py
"""
SIMULATION
*magic, do not touch*
"""

import pandas as pd
import scanpy as sc
import anndata as an
import numpy as np
from random import sample
from typing import Union, Optional, List


class Simulator(object):
    """
    Down-stream
    handle scMultiSim outputs
    """

    def __init__(self):
        self.data = None
        self.tfs = None
        self.tf_names = None
        self.motif_names = None
        self.receptors = None
        self.ligands = None
        self.targets = None
        self.total_genes = None

        self.lr_gt = None
        self.grn_gt = None

        self.counts = None
        self.noise_ids = None  # noise genes
        self.real_ids = None  # real genes
        self.total_name = None
        self.total_id = None

        self.name_df = None

        self.coor = None
        self.celltypes = None

    def get_lr(self, df):
        """
        Load Ligand-Receptor ground truth if available
        :param df:
        :return:
        """
        self.lr_gt = df
        self.receptors = list(set(df.receptor))
        self.ligands = list(set(df.ligand))

    def get_tg(self, df):
        """
        Load TF-Target ground truth (must have)
        :param df:
        :return:
        """
        self.grn_gt = df
        self.targets = list(set(df['regulated.gene']))
        self.tfs = list(set(df['regulator.gene']))

    def recognize_ids(self):
        """
        for counts gene index
        :return:
        """
        if self.counts:
            self.total_genes = list(self.counts.index)
            self.noise_ids = [i for i in self.total_genes if 'gene' in i]
            self.real_ids = [i for i in self.total_genes if 'gene' not in i]
        else:
            raise ValueError("Expression matrix is not available, load expression matrix first")

    def add_coor(self, value: Union[str, pd.DataFrame]):
        if isinstance(value, pd.DataFrame):
            self.coor = value
        elif isinstance(value, str):
            self.coor = pd.read_csv(value, index_col=0)

    def add_celltypes(self, value: Union[str, pd.DataFrame]):
        if isinstance(value, pd.DataFrame):
            self.celltypes = value
        elif isinstance(value, str):
            self.celltypes = pd.read_csv(value, index_col=0)

    def load_exp(self, df: pd.DataFrame):
        """
        when loading only one csv file created by scMultiSim
        :param df:
        :return:
        """
        self.counts = df
        self.recognize_ids()

    def load_multi_samples(self):
        """
        when loading multiple csv files generated by scMultiSim
        :return:
        """
        # Load in needed data
        df_list = []
        n = 0
        for num, tf in enumerate(self.tfs[:5]):
            df2 = pd.read_csv(f'counts_{tf}.csv', index_col=0)
            new_cell_num = df2.shape[1]
            new_cell_names = [f'cell{x}' for x in list(range(n + 1, n + new_cell_num + 1))]
            df2.columns = new_cell_names
            n = n + new_cell_num
            df_list.append(df2)
        # merge into one dataframe
        df = pd.concat(df_list, axis=1).fillna(0).astype(int)
        df3 = pd.read_csv('counts_addition.csv', index_col=0)
        gene_index = [i if 'gene' not in i else f'{i}3' for i in list(df3.index)]
        df3.index = gene_index
        df = pd.concat([df, df3]).fillna(0).astype(int)
        # self.noise_ids = [i for i in list(df.index) if 'gene' in i]
        self.counts = df
        self.recognize_ids()
        return df

    @staticmethod
    def load_multi(lists):
        """
        when loading multiple csv files generated by scMultiSim
        :return:
        """
        # Load in needed data
        df_list = []
        n = 0
        for num, i in enumerate(lists):
            df2 = pd.read_csv(f'counts_{i}.csv', index_col=0)
            new_cell_num = df2.shape[1]
            new_cell_names = [f'cell{x}' for x in list(range(n + 1, n + new_cell_num + 1))]
            df2.columns = new_cell_names
            n = n + new_cell_num
            df_list.append(df2)
        # merge into one dataframe
        df = pd.concat(df_list, axis=1).fillna(0).astype(int)
        df3 = pd.read_csv('counts_addition.csv', index_col=0)
        gene_index = [i if 'gene' not in i else f'{i}3' for i in list(df3.index)]
        df3.index = gene_index
        df = pd.concat([df, df3]).fillna(0).astype(int)
        return df

    def load_multiple_gts(self):
        """
        multiple GRN gts
        :return:
        """
        pass

    @staticmethod
    def get_dir(gt: pd.DataFrame, regulators, regulator_key='regulator.gene', value_key='regulator.effect',
                regulated_key='regulated.gene'):
        """
        Create dictionaries to map TF names to motif names and TFs to TF names
        :param gt:
        :param regulators: regulator: TF or ligand
        :param regulator_key:
        :param value_key:
        :param regulated_key:
        :return:
        """
        dir = {}
        for reg in regulators:
            sub = gt[gt[regulator_key] == reg]
            sorted_sub = sub.sort_values(by=value_key, ascending=False)
            dir[reg] = list(sorted_sub[regulated_key])
        return dir

    def assign_gene_names(self, rdb: pd.DataFrame, tf_id_dir, tf_motif_dir):
        grn_dir = Simulator.get_dir(self.grn_gt, self.tfs)
        # 1. for targets:
        # tf targets names, assign real names to tf targets. choose top n genes for each motif(TF).
        for tf in self.tfs:
            current_len = len(self.total_name)  # find current total_name number
            one_motif = tf_motif_dir[tf_id_dir[tf]]  # for each motif, find its gene rankings
            # find top genes for the motif
            sub = rdb[rdb.motifs == one_motif].drop('motifs', axis=1)  # remove motif column, only sort gene columns
            sorted_sub = sub.sort_values(by=(sub.index.values[0]), axis=1)  # sort columns (genes) by rank numbers
            top_num = len(grn_dir[tf])
            self.total_id += grn_dir[tf]
            # match real gene name with gene id
            for tg in sorted_sub.columns:
                if tg not in self.total_name:
                    self.total_name.append(tg)
                if len(self.total_name) == (current_len + top_num):
                    break

        # at this point, total names are unique, total ids have duplicates
        # remove duplicates
        name_df1 = pd.DataFrame({'id': self.total_id,
                                 'name': list(self.total_name)}).drop_duplicates(subset='id', keep='first').astype(str)
        self.total_name = list(name_df1.name)
        self.total_id = list(name_df1.id)
        # at this point, len(set(total_name)) should equal to len(set(total_id))
        assert name_df1.duplicated().sum().sum() == 0
        self.name_df = name_df1
        return name_df1

    def assign_name_addition(self, rdb, gene_id_list: list):
        """
        based on rest_rdb_names, when rest_rdb_names =
        :param rdb:
        :param gene_id_list: e.g noise_ids
        :return:
        """
        # PART TWO
        # some of the gene names are used. find out the rest unused gene names,
        # to avoid assign used gene name to noise genes and ligand and receptors
        total_rdb_names = set(rdb.columns) - set('motifs')
        rest_rdb_names = list(total_rdb_names - set(self.total_name))
        # 2. for noise genes
        noise_genes_names = sample(rest_rdb_names, k=len(gene_id_list))
        # !! random.choices returns duplicated items, use random.sample instead
        # add noise names into total_name
        self.total_name += noise_genes_names
        self.total_id += gene_id_list

        # create names df
        name_df = pd.DataFrame({'id': self.total_id,
                                'name': list(self.total_name)}).drop_duplicates(subset='id', keep='first').astype(str)
        assert name_df.duplicated().sum().sum() == 0
        self.name_df = name_df
        return name_df

    def to_anndata(self):
        """

        :param celltypes: cell types
        :param coor: cell coordinates
        :param name_df: simulated ID - gene name
        :return: anndata.Anndata
        """
        df = self.counts.T
        adata = sc.AnnData(df)
        adata.obs['cells'] = list(df.index)
        adata.var['genes'] = list(df.columns)
        if self.celltypes:
            adata.obs['celltype'] = self.celltypes['cell.type']
        if self.coor:
            adata.obsm['spatial'] = self.coor.iloc[:int(df.shape[0])]
        data_genes = adata.var.copy()
        data_genes = data_genes['genes'].replace(list(self.name_df['id']), list(self.name_df['name']))
        data_genes = data_genes.to_frame()
        adata.var = data_genes
        adata.var_names = adata.var['genes']
        return adata

    @staticmethod
    def subset(adata: an.AnnData, n_samples: int = 100, label: str = 'celltype',
               sub_clusters: Optional[list] = None) -> an.AnnData:
        """
        Subset samples
        :param adata:
        :param n_samples: number of samples to keep. should be smaller than total sample size.
        :param label:
        :param sub_clusters:
        :return:

        Example:
            subset(adata, n_samples=300, label='annotation', sub_clusters=['ct1','ct2']
        """
        l = []
        if sub_clusters:
            adata_cluster_sampled = adata[adata.obs[label].isin(sub_clusters)].copy()
        else:
            adata_cluster_sampled = adata.copy()

        for k, v in adata_cluster_sampled.obs.groupby(label).indices.items():
            l.append(np.random.choice(v, n_samples, replace=False))

        return adata_cluster_sampled[np.concatenate(l)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@file: regulatory_network.py
@time: 2023/Jan/08
@description: infer gene regulatory networks
@author: Yao LI
@email: liyao1@genomics.cn
@last modified by: Yao LI
"""

# python core modules
import os

# third party modules
import warnings
import json
import glob
import anndata
import hotspot
import pickle
import scipy
import pandas as pd
from copy import deepcopy
from multiprocessing import cpu_count
from typing import Sequence, Type, Optional, Optional, List

from dask.diagnostics import ProgressBar
from dask.distributed import Client, LocalCluster

from ctxcore.genesig import Regulon, GeneSignature
from arboreto.algo import grnboost2
from ctxcore.rnkdb import FeatherRankingDatabase as RankingDatabase
from pyscenic.utils import modules_from_adjacencies
from pyscenic.aucell import aucell, derive_auc_threshold
from pyscenic.prune import prune2df, df2regulons

# modules in self project
from .network import Network
from .autocor import *
from .corexp import *
from .c_autocor import gearys_c
from .m_autocor import morans_i_p_values, morans_i_zscore
from .g_autocor import getis_g


def intersection_ci(iterableA, iterableB, key=lambda x: x) -> list:
    """
    Return the intersection of two iterables with respect to `key` function.
    (ci: case insensitive)
    :param iterableA: list no.1
    :param iterableB: list no.2
    :param key:
    :return:
    """

    def unify(iterable):
        d = {}
        for item in iterable:
            d.setdefault(key(item), []).append(item)
        return d

    A, B = unify(iterableA), unify(iterableB)
    matched = []
    for k in A:
        if k in B:
            matched.append(B[k][0])
    return matched


def _name(fname: str) -> str:
    """
    Extract file name (without path and extension)
    :param fname:
    :return:
    """
    return os.path.splitext(os.path.basename(fname))[0]


def _set_client(num_workers: int) -> Client:
    """
    set number of processes when perform parallel computing
    :param num_workers:
    :return:
    """
    local_cluster = LocalCluster(n_workers=num_workers, threads_per_worker=1)
    custom_client = Client(local_cluster)
    return custom_client


def save_list(l, fn='list.txt'):
    """save a list into a text file"""
    with open(fn, 'w') as f:
        f.write('\n'.join(l))


class InferNetwork(Network):
    """
    Algorithms to infer Gene Regulatory Networks (GRNs)
    """

    def __init__(self, adata=None, project_name: str = 'project', ):
        """
        Constructor of this Object.
        :param data: sequencing data in AnnData format
        :return:
        """
        super().__init__()
        self.data = adata
        self.project_name = project_name

        self.more_stats = None
        self.weights = None
        self.ind = None
        self.weights_n = None

        # other settings
        self._params = {
            'rank_threshold': 1500,
            'prune_auc_threshold': 0.05,
            'nes_threshold': 3.0,
            'motif_similarity_fdr': 0.05,
            'auc_threshold': 0.05,
            'noweights': False,
        }
        self.tmp_dir = None

    # GRN pipeline infer logic
    def infer(self,
              databases: str,
              motif_anno_fn: str,
              tfs_fn,
              gene_list: Optional[List] = None,
              cluster_label='annotation',
              niche_df=None,
              receptor_key='to',
              num_workers=None,
              save_tmp=False,
              cache=False,

              layers='raw_counts',
              model='bernoulli',
              latent_obsm_key='spatial',
              umi_counts_obs_key=None,
              n_neighbors=30,
              weighted_graph=False,
              rho_mask_dropouts=False,
              local=False,
              methods=None,
              operation='intersection',
              combine=False,
              mode='moran',
              somde_k=20,
              noweights=None,
              normalize: bool = False):
        if save_tmp:
            self.tmp_dir = os.path.join(self.project_name, 'tmp_files')
            if not os.path.exists(self.tmp_dir):
                os.makedirs(self.tmp_dir)

        global adjacencies
        exp_mat = self._data.to_df()

        if num_workers is None:
            num_workers = cpu_count()

        if noweights is None:
            noweights = self.params["noweights"]

        # 1. load TF list
        if tfs_fn is None:
            tfs = 'all'
        else:
            tfs = self.load_tfs(tfs_fn)

        # 2. load the ranking databases
        dbs = self.load_database(databases)

        # 3. GRN Inference
        adjacencies = self.spg(self.data,
                               gene_list=gene_list,
                               tf_list=tfs,
                               jobs=num_workers,
                               layer_key=layers,
                               model=model,
                               latent_obsm_key=latent_obsm_key,
                               umi_counts_obs_key=umi_counts_obs_key,
                               n_neighbors=n_neighbors,
                               weighted_graph=weighted_graph,
                               cache=cache,
                               save_tmp=save_tmp,
                               fn=f'{self.project_name}_{mode}_adj.csv',
                               local=local,
                               methods=methods,
                               operation=operation,
                               combine=combine,
                               mode=mode,
                               somde_k=somde_k)

        # 4. Compute Modules
        # ctxcore.genesig.Regulon
        modules = self.get_modules(adjacencies,
                                   exp_mat,
                                   cache=cace,
                                   save_tmp=save_tmp,
                                   rho_mask_dropouts=rho_mask_dropouts)
        # before_cistarget(tfs, modules, project_name)

        # 5. Regulons Prediction aka cisTarget
        # ctxcore.genesig.Regulon
        regulons = self.prune_modules(modules,
                                      dbs,
                                      motif_anno_fn,
                                      num_workers=num_workers,
                                      save_tmp=save_tmp,
                                      cache=cache,
                                      fn=f'{self.project_name}_motifs.csv',
                                      prefix=self.project_name,
                                      rank_threshold=self.params["rank_threshold"],
                                      auc_threshold=self.params["prune_auc_threshold"],
                                      nes_threshold=self.params["nes_threshold"],
                                      motif_similarity_fdr=self.params["motif_similarity_fdr"])

        # 6.0. Cellular Enrichment (aka AUCell)
        self.cal_auc(exp_mat,
                     regulons,
                     auc_threshold=self.params["auc_threshold"],
                     num_workers=num_workers,
                     save_tmp=save_tmp,
                     cache=cache,
                     noweights=noweights,
                     normalize=normalize,
                     fn=f'{self.tmp_dir}/auc.csv')
        # self.compute_regulons(adjacencies,
        #                       exp_mat,
        #                       dbs,
        #                       motif_anno_fn,
        #                       rho_mask_dropouts=rho_mask_dropouts,
        #                       num_workers=num_workers,
        #                       save_tmp=save_tmp,
        #                       cache=cache,
        #                       noweights=noweights,
        #                       normalize=normalize)

        # 6.1. Receptor AUCs
        if niche_df is not None:
            self.get_filtered_receptors(niche_df, receptor_key=receptor_key)
            receptor_auc_mtx = self.receptor_auc()
            self.isr(receptor_auc_mtx)

        # 7. Calculate Regulon Specificity Scores
        self.cal_regulon_score(cluster_label=cluster_label, save_tmp=save_tmp,
                               fn=f'{self.tmp_dir}/regulon_specificity_scores.txt')

        # 8. Save results to h5ad file
        # TODO: check if data has adj, regulon_dict, auc_mtx etc. before saving to disk
        # dtype=object
        self.data.write_h5ad(f'{self.project_name}_spagrn.h5ad')
        return self.data

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, value):
        """only use this function when setting params as a whole.
        use add_params to solely update/add some of the params and keep the rest unchanged"""
        self._params = value

    def add_params(self, dic: dict):
        """
        :param dic: keys are parameter name, values are parameter values

        Example:
            grn = InferNetwork(data)
            grn.add_params({'num_worker':12, 'auc_threshold': 0.001})
        """
        og_params = deepcopy(self._params)
        try:
            for key, value in dic.items():
                self._params[key] = value
        except KeyError:
            self._params = og_params

    # ------------------------------------------------------#
    #             step0: LOAD AUXILIARY DATA                #
    # ------------------------------------------------------#
    @staticmethod
    def read_motif_file(fname):
        """
        Read motifs.csv file generate by
        :param fname:
        :return:
        """
        df = pd.read_csv(fname, sep=',', index_col=[0, 1], header=[0, 1], skipinitialspace=True)
        df[('Enrichment', 'Context')] = df[('Enrichment', 'Context')].apply(lambda s: eval(s))
        df[('Enrichment', 'TargetGenes')] = df[('Enrichment', 'TargetGenes')].apply(lambda s: eval(s))
        return df

    @staticmethod
    def load_tfs(fn: str) -> list:
        """
        Get a list of interested TFs from a text file
        :param fn:
        :return:
        """
        with open(fn) as file:
            tfs_in_file = [line.strip() for line in file.readlines()]
        return tfs_in_file

    @staticmethod
    def load_database(database_dir: str) -> list:
        """
        Load motif ranking database
        :param database_dir:
        :return:
        """
        db_fnames = glob.glob(database_dir)
        dbs = [RankingDatabase(fname=fname, name=_name(fname)) for fname in db_fnames]
        return dbs

    # ------------------------------------------------------#
    #           step1: CALCULATE TF-GENE PAIRS              #
    # ------------------------------------------------------#
    def rf_infer(self,
                 matrix,
                 tf_names,
                 genes: list,
                 num_workers: int,
                 verbose: bool = True,
                 cache: bool = True,
                 save_tmp: bool = True,
                 fn: str = 'adj.csv',
                 **kwargs) -> pd.DataFrame:
        """
        Inference of co-expression modules via random forest (RF) module
        :param matrix:
            * pandas DataFrame (rows=observations, columns=genes)
            * dense 2D numpy.ndarray
            * sparse scipy.sparse.csc_matrix
        :param tf_names: list of target TFs or all
        :param genes: list of interested genes
        :param num_workers: number of thread
        :param verbose: if print out running details
        :param cache:
        :param save_tmp: if save adjacencies result into a file
        :param fn: adjacencies file name
        :return:
        """
        if cache and os.path.isfile(fn):
            adjacencies = pd.read_csv(fn)
            self.adjacencies = adjacencies
            self.data.uns['adj'] = adjacencies
            return adjacencies

        if num_workers is None:
            num_workers = cpu_count()
        custom_client = _set_client(num_workers)
        adjacencies = grnboost2(matrix,
                                tf_names=tf_names,
                                gene_names=genes,
                                verbose=verbose,
                                client_or_address=custom_client,
                                **kwargs)
        if save_tmp:
            adjacencies.to_csv(os.path.join(self.tmp_dir, fn), index=False)
        self.adjacencies = adjacencies
        self.data.uns['adj'] = adjacencies
        return adjacencies

    def spatial_autocorrelation(self,
                                adata,
                                layer_key="raw_counts",
                                latent_obsm_key="spatial",
                                n_neighbors=10,
                                somde_k=20,
                                n_processes=None,
                                local=False,
                                cache=False):
        """
        Calculate spatial autocorrelation values using Moran's I, Geary'C, Getis's G and SOMDE algorithms
        :param adata:
        :param layer_key:
        :param latent_obsm_key:
        :param n_neighbors:
        :param somde_k:
        :param n_processes:
        :param local:
        :param cache:
        :return:
        """
        print('Computing spatial weights matrix...')
        self.ind, neighbors, self.weights_n = neighbors_and_weights(adata, latent_obsm_key=latent_obsm_key,
                                                                    n_neighbors=n_neighbors)
        Weights = get_w(self.ind, self.weights_n)
        self.weights = Weights

        more_stats = pd.DataFrame(index=adata.var_names)
        if local:
            if cache and os.path.isfile('local_more_stats.csv'):
                print('Found file local_more_stats.csv')
                more_stats = pd.read_csv('local_more_stats.csv', index_col=0, sep='\t')
                self.more_stats = more_stats
                return more_stats
            print('Computing SOMDE...')
            adjusted_p_values = somde_p_values(adata, k=somde_k, layer_key=layer_key, latent_obsm_key=latent_obsm_key)
            more_stats['FDR_SOMDE'] = adjusted_p_values
            more_stats.to_csv('local_more_stats.csv', sep='\t')
        else:
            if cache and os.path.isfile('more_stats.csv'):
                print('Found file more_stats.csv')
                more_stats = pd.read_csv('more_stats.csv', index_col=0, sep='\t')
                self.more_stats = more_stats
                return more_stats
            print("Computing Moran's I...")
            morans_ps = morans_i_p_values(adata, Weights, layer_key=layer_key, n_process=n_processes)
            fdr_morans_ps = fdr(morans_ps)
            print("Computing Geary's C...")
            gearys_cs = gearys_c(adata, Weights, layer_key=layer_key, n_process=n_processes, mode='pvalue')
            fdr_gearys_cs = fdr(gearys_cs)
            print("Computing Getis G...")
            getis_gs = getis_g(adata, Weights, n_processes=n_processes, layer_key=layer_key, mode='pvalue')
            fdr_getis_gs = fdr(getis_gs)
            # save results
            more_stats = pd.DataFrame({
                'C': gearys_cs,
                'FDR_C': fdr_gearys_cs,
                'I': morans_ps,
                'FDR_I': fdr_morans_ps,
                'G': getis_gs,
                'FDR_G': fdr_getis_gs
            }, index=adata.var_names)

        self.more_stats = more_stats
        return more_stats

    @staticmethod
    def spatial_autocorrelation_zscore(adata,
                                       layer_key="raw_counts",
                                       latent_obsm_key="spatial",
                                       n_neighbors=10,
                                       n_processes=None):
        """
        Calculate spatial autocorrelation values using Moran's I, Geary'C, Getis's G and SOMDE algorithms
        :param adata:
        :param layer_key:
        :param latent_obsm_key:
        :param n_neighbors:
        :param n_processes:
        :return:
        """
        print('Computing spatial weights matrix...')
        ind, neighbors, weights_n = neighbors_and_weights(adata, latent_obsm_key=latent_obsm_key,
                                                          n_neighbors=n_neighbors)
        Weights = get_w(ind, weights_n)

        print("Computing Moran's I...")
        morans_ps = morans_i_zscore(adata, Weights, layer_key=layer_key, n_process=n_processes)
        print("Computing Geary's C...")
        gearys_cs = gearys_c(adata, Weights, layer_key=layer_key, n_process=n_processes, mode='zscore')
        print("Computing Getis G...")
        getis_gs = getis_g(adata, Weights, n_processes=n_processes, layer_key=layer_key, mode='zscore')
        # save results
        more_stats = pd.DataFrame({
            'C_zscore': gearys_cs,
            'I_zscore': morans_ps,
            'G_zscore': getis_gs,
        }, index=adata.var_names)
        return more_stats

    def select_genes(self, methods=None, fdr_threshold=0.05, local=True, combine=True, operation='intersection'):
        """
        Select genes based FDR values...
        :param methods:
        :param fdr_threshold:
        :param local:
        :param combine:
        :param operation:
        :return:
        """
        if methods is None:
            methods = ['FDR_C', 'FDR_I', 'FDR_G', 'FDR']
        # 1. LOCAL
        if local:
            somde_genes = self.more_stats.loc[self.more_stats.FDR_SOMDE < fdr_threshold].index
            print(f'SOMDE find {len(somde_genes)} genes')
            return somde_genes
        # 2. GLOBAL
        # 2.1 combine p-values
        elif combine:
            cfdrs = combind_fdrs(self.more_stats[['FDR_C', 'FDR_I', 'FDR_G', 'FDR']])  # combine 4 types of p-values
            self.more_stats['combined'] = cfdrs
            genes = self.more_stats.loc[self.more_stats['combined'] < fdr_threshold].index
            print(f"Combinded FDRs gives: {len(cgenes)} genes")
            return genes
        # 2.2 individual p-values
        elif methods:
            indices_list = [set(self.more_stats[self.more_stats[m] < fdr_threshold].index) for m in methods]
            if operation == 'intersection':
                global_inter_genes = set.intersection(*indices_list)
                print(f'global spatial gene num (intersection): {len(global_inter_genes)}')
                return global_inter_genes
            elif operation == 'union':
                global_union_genes = set().union(*indices_list)
                print(f'global spatial gene num (union): {len(global_union_genes)}')
                return global_union_genes

    @staticmethod
    def check_stats(more_stats):
        """Compute gene numbers for each"""
        moran_genes = more_stats.loc[more_stats.FDR_I < fdr_threshold].index
        geary_genes = more_stats.loc[more_stats.FDR_C < fdr_threshold].index
        getis_genes = more_stats.loc[more_stats.FDR_G < fdr_threshold].index
        hs_genes = more_stats.loc[(more_stats.FDR < fdr_threshold)].index
        print(f"Moran's I find {len(moran_genes)} genes")
        print(f"Geary's C find {len(geary_genes)} genes")
        print(f'Getis find {len(getis_genes)} genes')
        print(f"HOTSPOT find {len(hs_genes)} genes")
        if 'FDR_SOMDE' in more_stats.columns:
            somde_genes = more_stats.loc[more_stats.FDR_SOMDE < fdr_threshold].index
            print(f'SOMDE find {len(somde_genes)} genes')

    def spg(self,
            data: anndata.AnnData,
            gene_list: Optional[List] = None,
            layer_key=None,
            model='bernoulli',
            latent_obsm_key="spatial",
            umi_counts_obs_key=None,
            weighted_graph=False,
            n_neighbors=10,
            fdr_threshold=0.05,
            tf_list=None,
            save_tmp=False,
            jobs=None,
            cache=False,
            local=False,
            methods=None,
            operation='intersection',
            combine=True,
            somde_k=20,
            fn: str = 'adj.csv',
            mode='moran',
            **kwargs) -> pd.DataFrame:
        """
        Inference of co-expression modules by spatial-proximity-graph (SPG) model.
        :param data: Count matrix (shape is cells by genes)
        :param gene_list: A list of interested genes to calculate co-expression values with TF genes.
                Could be HVGs or all genes in the count matrix. When not provided, will compute spatial autocorrelation
                values between all genes in the count matrix with TF genes and select interested genes with significant
                spatial variability.
        :param layer_key: Key in adata.layers with count data, uses adata.X if None.
        :param model: Specifies the null model to use for gene expression.
            Valid choices are:
                * 'danb': Depth-Adjusted Negative Binomial
                * 'bernoulli': Models probability of detection
                * 'normal': Depth-Adjusted Normal
                * 'none': Assumes data has been pre-standardized
        :param latent_obsm_key: Latent space encoding cell-cell similarities with euclidean
            distances.  Shape is (cells x dims). Input is key in adata.obsm
        :param umi_counts_obs_key: Total umi count per cell.  Used as a size factor.
            If omitted, the sum over genes in the counts matrix is used. 'total_counts'
        :param weighted_graph: Whether or not to create a weighted graph
        :param n_neighbors: Neighborhood size
        :param fdr_threshold: Correlation threshold at which to stop assigning genes to modules
        :param tf_list: predefined TF names
        :param save_tmp: if save results onto disk
        :param jobs: Number of parallel jobs to run_all
        :param cache:
        :param local:
        :param combine:
        :param mode:
        :param somde_k:
        :param operation:
        :param methods:
        :param fn: output file name
        :return: A dataframe, local correlation Z-scores between genes (shape is genes x genes)
        """
        global local_correlations
        if cache and os.path.isfile(fn):
            print(f'Found file {fn}')
            local_correlations = pd.read_csv(fn)
            self.data.uns['adj'] = local_correlations
            return local_correlations
        else:
            # 2024-12-20: select genes or provide a list of genes
            if gene_list:
                hs_genes = gene_list
                print('Computing spatial weights matrix...')
                self.ind, neighbors, self.weights_n = neighbors_and_weights(data, latent_obsm_key=latent_obsm_key,
                                                                            n_neighbors=n_neighbors)
                Weights = get_w(self.ind, self.weights_n)
                self.weights = Weights
            else:
                hs = hotspot.Hotspot(data,
                                     layer_key=layer_key,
                                     model=model,
                                     latent_obsm_key=latent_obsm_key,
                                     umi_counts_obs_key=umi_counts_obs_key,
                                     **kwargs)
                hs.create_knn_graph(weighted_graph=weighted_graph, n_neighbors=n_neighbors)
                hs_results = hs.compute_autocorrelations()

                # 1: Select genes
                self.spatial_autocorrelation(data,
                                             layer_key=layer_key,
                                             latent_obsm_key=latent_obsm_key,
                                             n_neighbors=n_neighbors,
                                             somde_k=somde_k,
                                             n_processes=jobs,
                                             local=local,
                                             cache=cache)
                self.more_stats['FDR'] = hs_results.FDR
                if save_tmp:
                    self.more_stats.to_csv(f'{self.tmp_dir}/more_stats.csv', sep='\t')

                hs_genes = self.select_genes(methods=methods,
                                             fdr_threshold=fdr_threshold,
                                             local=local,
                                             combine=combine,
                                             operation=operation)
                hs_genes = list(hs_genes)
                assert len(hs_genes) > 0
                if save_tmp:
                    save_list(hs_genes, fn=f'{self.tmp_dir}/selected_genes.txt')

            # 2. Define gene-gene relationships with pair-wise local correlations
            print(f'Current mode is {mode}')
            if mode == 'zscore':
                # subset by TFs
                local_correlations = hs.compute_local_correlations(hs_genes, jobs=jobs)  # jobs for parallelization
                if tf_list:
                    common_tf_list = list(set(tf_list).intersection(set(local_correlations.columns)))
                    assert len(common_tf_list) > 0, 'predefined TFs not found in data'
                else:
                    common_tf_list = local_correlations.columns
                # reshape matrix
                local_correlations['TF'] = local_correlations.columns
                local_correlations = local_correlations.melt(id_vars=['TF'])
                local_correlations.columns = ['TF', 'target', 'importance']
                local_correlations = local_correlations[local_correlations.TF.isin(common_tf_list)]
                # remove if TF = target
                local_correlations = local_correlations[local_correlations.TF != local_correlations.target]

            elif mode == 'moran':
                tfs_in_data = list(set(tf_list).intersection(set(data.var_names)))
                select_genes_not_tfs = list(set(hs_genes) - set(tfs_in_data))
                fw = flat_weights(data.obs_names, self.ind, self.weights_n, n_neighbors=n_neighbors)
                local_correlations = global_bivariate_moran_R(data,
                                                              fw,
                                                              tfs_in_data,
                                                              select_genes_not_tfs,
                                                              num_workers=jobs,
                                                              layer_key=layer_key)

            elif mode == 'geary':
                tfs_in_data = list(set(tf_list).intersection(set(data.var_names)))
                select_genes_not_tfs = list(set(hs_genes) - set(tfs_in_data))
                fw = flat_weights(data.obs_names, self.ind, self.weights_n, n_neighbors=n_neighbors)
                local_correlations = global_bivariate_gearys_C(data,
                                                               fw,
                                                               tfs_in_data,
                                                               select_genes_not_tfs,
                                                               num_workers=jobs,
                                                               layer_key=layer_key)
        # 2025-01-05
        local_correlations['importance'] = local_correlations['importance'].astype(np.float64)
        self.data.uns['adj'] = local_correlations
        if save_tmp:
            local_correlations.to_csv(os.path.join(self.tmp_dir, 'local_correlations.csv'), index=False)
        return local_correlations

    # ------------------------------------------------------#
    #            step2:  FILTER TFS AND TARGETS             #
    # ------------------------------------------------------#
    def get_modules(self,
                    adjacencies: pd.DataFrame,
                    matrix,
                    rho_mask_dropouts: bool = False,
                    save_tmp=False,
                    cache=False,
                    **kwargs) -> Sequence[Regulon]:
        """
        Create of co-expression modules
        :param adjacencies:
        :param matrix:
            * pandas DataFrame (rows=observations, columns=genes)
            * dense 2D numpy.ndarray
            * sparse scipy.sparse.csc_matrix
        :param rho_mask_dropouts:
        :return:
        """
        if cache and os.path.isfile(f'{self.tmp_dir}/modules.pkl'):
            modules = pickle.load(open(f'{self.tmp_dir}/modules.pkl', 'rb'))
            return modules
        modules = list(
            modules_from_adjacencies(adjacencies, matrix, rho_mask_dropouts=rho_mask_dropouts, **kwargs)
        )
        self.modules = modules
        if save_tmp:
            with open(f'{self.tmp_dir}/modules.pkl', "wb") as f:
                pickle.dump(modules, f)
        return modules

    # ------------------------------------------------------#
    #            step3:  FILTER TFS AND TARGETS             #
    # ------------------------------------------------------#
    def prune_modules(self,
                      modules: Sequence[Regulon],
                      dbs: list,
                      motif_anno_fn: str,
                      num_workers: int,
                      cache: bool = False,
                      save_tmp: bool = False,
                      fn: str = 'motifs.csv',
                      prefix: str = 'exp',
                      **kwargs) -> Sequence[Regulon]:
        """
        First, calculate a list of enriched motifs and the corresponding target genes for all modules.
        Then, create regulon_list from this table of enriched motifs.
        :param modules: The sequence of modules.
        :param dbs: The sequence of databases.
        :param motif_anno_fn: The name of the file that contains the motif annotations to use.
        :param rank_threshold: The total number of ranked genes to take into account when creating a recovery curve.
        :param auc_threshold: The fraction of the ranked genome to take into account for the calculation of the
            Area Under the recovery Curve.
        :param nes_threshold: The Normalized Enrichment Score (NES) threshold to select enriched features.
        :param motif_similarity_fdr: The maximum False Discovery Rate to find factor annotations for enriched motifs.
        :param orthologuous_identity_threshold: The minimum orthologuous identity to find factor annotations
            for enriched motifs.
        :param weighted_recovery: Use weights of a gene signature when calculating recovery curves?
        :param num_workers: If not using a cluster, the number of workers to use for the calculation.
            None of all available CPUs need to be used.
        :param module_chunksize: The size of the chunk to use when using the dask framework.
        :param cache:
        :param save_tmp:
        :param fn:
        :param prefix:
        :param kwargs:
        :return: A dataframe.
        """
        if cache and os.path.isfile(fn):
            df = self.read_motif_file(fn)
            regulon_list = df2regulons(df)
            self.regulons = regulon_list
            self.regulon_dict = self.get_regulon_dict(regulon_list)
            self.data.uns['regulon_dict'] = self.regulon_dict
            return regulon_list

        if num_workers is None:
            num_workers = cpu_count()
        # infer function
        # #1.
        with ProgressBar():
            df = prune2df(dbs, modules, motif_anno_fn, num_workers=num_workers, **kwargs)  # rank_threshold

        # this function actually did two things. 1. get df, 2. turn df into list of Regulons
        # #2.
        regulon_list = df2regulons(df)
        self.regulons = regulon_list

        # #3. handle results
        # convert Regulon list to dictionaries for easy access and readability
        self.regulon_dict = self.get_regulon_dict(regulon_list)
        self.data.uns['regulon_dict'] = self.regulon_dict

        # save to data
        if save_tmp:
            with open(f'{prefix}_regulons.json', 'w') as f:
                json.dump(self.regulon_dict, f, sort_keys=True, indent=4)
        return regulon_list

    # ------------------------------------------------------#
    #           step4: CALCULATE TF-GENE PAIRS              #
    # ------------------------------------------------------#
    def cal_auc(self,
                matrix,
                regulons: Sequence[Type[GeneSignature]],
                auc_threshold: float,
                num_workers: int,
                noweights: bool = False,
                normalize: bool = False,
                seed=None,
                cache: bool = True,
                save_tmp: bool = True,
                fn='auc.csv') -> pd.DataFrame:
        """
        Calculate enrichment of gene signatures for cells/spots.

        :param matrix:
            * pandas DataFrame (rows=observations, columns=genes)
            * dense 2D numpy.ndarray
            * sparse scipy.sparse.csc_matrix
        :param regulons: list of ctxcore.genesig.Regulon objects
        :param auc_threshold: The fraction of the ranked genome to take into account for the calculation of the
            Area Under the recovery Curve.
        :param num_workers: The number of cores to use.
        :param noweights: Should the weights of the genes part of a signature be used in calculation of enrichment?
        :param normalize: Normalize the AUC values to a maximum of 1.0 per regulon.
        :param seed: seed for generating random numbers
        :param cache:
        :param save_tmp:
        :param fn:
        :return: A dataframe with the AUCs (n_cells x n_modules).
        """
        if cache and os.path.isfile(fn):
            auc_mtx = pd.read_csv(fn, index_col=0)  # important! cell must be index, not one of the column
            self.auc_mtx = auc_mtx
            self.data.obsm['auc_mtx'] = self.auc_mtx
            return auc_mtx

        if num_workers is None:
            num_workers = cpu_count()

        auc_mtx = aucell(matrix,
                         regulons,
                         auc_threshold=auc_threshold,
                         num_workers=num_workers,
                         noweights=noweights,
                         normalize=normalize,
                         seed=seed)

        self.auc_mtx = auc_mtx
        self.data.obsm['auc_mtx'] = self.auc_mtx
        if save_tmp:
            auc_mtx.to_csv(os.path.join(self.tmp_dir, 'auc_mtx.csv'))
        return auc_mtx

    def receptor_auc(self, auc_threshold=None, p_range=0.01, num_workers=20) -> Optional[pd.DataFrame]:
        """
        Calculate AUC value for modules that detected receptor genes within
        :param auc_threshold:
        :param p_range:
        :param num_workers:
        :return:
        """
        if self.receptor_dict is None:
            print('receptor dict not found. run_all get_receptors first.')
            return
        # 1. create new modules
        receptor_modules = list(
            map(
                lambda x: GeneSignature(
                    name=x,
                    gene2weight=self.receptor_dict[x],
                ),
                self.receptor_dict,
            )
        )
        ex_matrix = self.data.to_df()
        if auc_threshold is None:
            percentiles = derive_auc_threshold(ex_matrix)
            a_value = percentiles[p_range]
        else:
            a_value = auc_threshold
        receptor_auc_mtx = aucell(ex_matrix, receptor_modules, auc_threshold=a_value, num_workers=num_workers)
        return receptor_auc_mtx

    def isr(self, receptor_auc_mtx) -> pd.DataFrame:
        """
        Calculate ISR matrix for all the regulons
        :param receptor_auc_mtx: auc matrix for modules containing receptor genes
        :return:
        """
        # receptor_auc_mtx = self.data.obsm['rep_auc_mtx']
        auc_mtx = self.data.obsm['auc_mtx']
        # change receptor auc matrix column names so it can aligned with the auc matrix column names
        col_names = receptor_auc_mtx.columns.copy()
        col_names = [f'{i}(+)' for i in col_names]
        receptor_auc_mtx.columns = col_names
        # ! only uses modules that occurs in the auc matrix aka they have been identified as regulons
        later_regulon_names = list(set(auc_mtx.columns).intersection(set(col_names)))
        receptor_auc_mtx = receptor_auc_mtx[later_regulon_names]
        # combine two dfs
        df = pd.concat([auc_mtx, receptor_auc_mtx], axis=1)
        # sum values
        isr_df = df.groupby(level=0, axis=1).sum()
        self.data.obsm['isr'] = isr_df
        return isr_df

    # ------------------------------------------------------ #
    #              step2-3: Receptors Detection              #
    # ------------------------------------------------------ #
    def get_filtered_genes(self):
        """
        Detect genes filtered by cisTarget
        :return:
        """
        module_tf = []
        for i in self.modules:
            module_tf.append(i.transcription_factor)

        final_tf = [i.strip('(+)') for i in list(self.regulon_dict.keys())]
        com = set(final_tf).intersection(set(module_tf))

        before_tf = {}
        for tf in com:
            before_tf[tf] = []
            for i in self.modules:
                if tf == i.transcription_factor:
                    before_tf[tf] += list(i.genes)

        filtered = {}
        for tf in com:
            final_targets = self.regulon_dict[f'{tf}(+)']
            before_targets = set(before_tf[tf])
            filtered_targets = before_targets - set(final_targets)
            if tf in filtered_targets:
                filtered_targets.remove(tf)
            filtered[tf] = list(filtered_targets)
        self.filtered = filtered
        self.data.uns['filtered_genes'] = filtered
        return filtered

    def get_filtered_receptors(self, niche_df: pd.DataFrame, receptor_key='to'):
        """

        :type niche_df: pd.DataFrame
        :param receptor_key: column name of receptor
        :param save_tmp:
        :param niche_df:
        :return:
        """
        if niche_df is None:
            warnings.warn("Ligand-Receptor reference database is missing, skipping get_filtered_receptors method")
            return

        receptor_tf = {}
        total_receptor = set()

        self.get_filtered_genes()
        for tf, targets in self.filtered.items():
            rtf = set(intersection_ci(set(niche_df[receptor_key]), set(targets), key=str.lower))
            if len(rtf) > 0:
                receptor_tf[tf] = list(rtf)
                total_receptor = total_receptor | rtf
        self.receptors = total_receptor
        self.receptor_dict = receptor_tf
        self.data.uns['receptor_dict'] = receptor_tf

    # ------------------------------------------------------------------------------------------------
    def compute_regulons(self, adjacencies, exp_mat, dbs, motif_anno_fn, rho_mask_dropouts, num_workers, save_tmp,
                         cache,
                         noweights, normalize):
        modules = self.get_modules(adjacencies, exp_mat, rho_mask_dropouts=rho_mask_dropouts, prefix=self.project_name)

        regulons = self.prune_modules(modules,
                                      dbs,
                                      motif_anno_fn,
                                      num_workers=num_workers,
                                      save_tmp=save_tmp,
                                      cache=cache,
                                      fn=f'{self.project_name}_motifs.csv',
                                      prefix=self.project_name,
                                      rank_threshold=self.params["rank_threshold"],
                                      auc_threshold=self.params["prune_auc_threshold"],
                                      nes_threshold=self.params["nes_threshold"],
                                      motif_similarity_fdr=self.params["motif_similarity_fdr"])

        self.cal_auc(exp_mat,
                     regulons,
                     auc_threshold=self.params["auc_threshold"],
                     num_workers=num_workers,
                     save_tmp=save_tmp,
                     cache=cache,
                     noweights=noweights,
                     normalize=normalize,
                     fn=f'{self.project_name}_auc.csv')



#!/usr/bin/env python
import setuptools
import glob
import os
from pathlib import Path

setuptools.setup(
    name="spagrn",
    version="1.1.21-beta",
    author="Yao LI",
    author_email="liyao1@genomics.cn",
    url="https://github.com/BGI-Qingdao/SpaGRN",
    #long_description=Path('README.md').read_text('utf-8'),
    python_requires=">=3.8,<3.9",
    packages=setuptools.find_packages(),
    install_requires=[
        #"anndata==0.8.0",
        "anndata>=0.8.0,<0.9",
        "pandas<2.0.0, >=1.3.4",
        #"scanpy==1.9.1",
        "scanpy>=1.9.1,<1.9.4",
        #"seaborn",
        "seaborn<0.13",
        "matplotlib<=3.5.3",
        "pyscenic==0.12.1",
        "hotspotsc==1.1.1",
        "scipy",
        "numpy<1.24,>=1.16.6",
        "dask",
        "arboreto",
        "ctxcore>=0.2.0",
        "scikit-learn",
        "esda"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    license= 'GPL-3.0-or-later',
    description="A comprehensive tool to infer TF-centered, spatial gene regulatory networks for the spatially resolved transcriptomic data.",
    platforms='any',
    entry_points={
        "console_scripts": [
            "spagrn = spagrn.cli.spagrn_parser:main",
        ],
    },
)




# !/usr/bin/env python -*- coding: utf-8 -*-
# @Date: Created on 31 Oct 2023 15:00
# @Author: Yao LI
# @File: spagrn/network.py
# @Description: A Gene Regulatory Network object. A typical network essentially contains regulators
# (e.g. TFs), target genes. and cell types, regulons score between cell types and activity level among cells. and
# regulator-target regulatory effect.

import os

import json
import pickle
import pandas as pd
import scanpy as sc
import numpy as np
import anndata as an
from typing import Sequence

from ctxcore.genesig import Regulon
from pyscenic.rss import regulon_specificity_scores


def remove_all_zero(auc_mtx):
    # check if there were regulons contain all zero auc values
    auc_mtx = auc_mtx.loc[:, ~auc_mtx.ne(0).any()]
    # remove all zero columns (which have no variation at all)
    auc_mtx = auc_mtx.loc[:, (auc_mtx != 0).any(axis=0)]
    return auc_mtx


class Network:
    def __init__(self):
        """
        Constructor of the (Gene Regulatory) Network Object.
        """
        # input
        self._data = None  # anndata.Anndata
        self._matrix = None  # pd.DataFrame
        self._gene_names = None  # list of strings
        self._cell_names = None  # list of strings
        self._position = None  # np.array
        self._tfs = None  # list

        # Network calculated attributes
        self._regulons = None  # list of ctxcore.genesig.Regulon instances
        self._modules = None  # list of ctxcore.genesig.Regulon instances
        self._auc_mtx = None  # pd.DataFrame
        self._adjacencies = None  # pd.DataFrame
        self._regulon_dict = None  # dictionary
        self._rss = None  # pd.DataFrame

        # Receptors
        self._filtered = None  # dictionary
        self._receptors = None  # set
        self.receptor_dict = None

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def matrix(self):
        return self._matrix

    @matrix.setter
    def matrix(self, value):
        self._matrix = value

    @property
    def gene_names(self):
        return self._gene_names

    @gene_names.setter
    def gene_names(self, value):
        self._gene_names = value

    @property
    def cell_names(self):
        return self._cell_names

    @cell_names.setter
    def cell_names(self, value):
        self._cell_names = value

    @property
    def adjacencies(self):
        return self._adjacencies

    @adjacencies.setter
    def adjacencies(self, value):
        self._adjacencies = value

    @property
    def regulons(self):
        return self._regulons

    @regulons.setter
    def regulons(self, value):
        self._regulons = value

    @property
    def regulon_dict(self):
        return self._regulon_dict

    @regulon_dict.setter
    def regulon_dict(self, value):
        self._regulon_dict = value

    @property
    def auc_mtx(self):
        return self._auc_mtx

    @auc_mtx.setter
    def auc_mtx(self, value):
        self._auc_mtx = value

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        self._position = value

    @property
    def rss(self):
        return self._rss

    @rss.setter
    def rss(self, value):
        self._rss = value

    @property
    def modules(self):
        return self._modules

    @modules.setter
    def modules(self, value):
        self._modules = value

    @property
    def filtered(self):
        return self._filtered

    @filtered.setter
    def filtered(self, value):
        self._filtered = value

    @property
    def receptors(self):
        return self._receptors

    @receptors.setter
    def receptors(self, value):
        self._receptors = value

    # ------------------------------------------------------#
    #                Data loading methods                   #
    # ------------------------------------------------------#
    def load_data_info(self, pos_label='spatial'):
        """
        (for raw data)
        Load useful data to properties.
        :param pos_label: pos key in obsm, default 'spatial'. Only used if data is Anndata.
        :return:
        """
        if self.data:
            self.matrix = self.data.X
            self.gene_names = self.data.var_names
            self.cell_names = self.data.obs_names
            self.position = self.data.obsm[pos_label]

    def load_results(self, modules_fn=None, regulons_fn=None):
        """
        (for derived data)
        Load results generate by SpaGRN. Mainly contains
        :param modules_fn:
        :param regulons_fn:
        :return:
        """
        try:
            self.regulon_dict = self.data.uns['regulon_dict']
            self.adjacencies = self.data.uns['adj']
            self.auc_mtx = self.data.obsm['auc_mtx']
            self.rss = self.data.uns['rss']
        except KeyError as e:
            print(f"WARNING: {e.args[0]} does not exist")
        if modules_fn:
            self.modules = pickle.load(open(modules_fn, 'rb'))
        if regulons_fn:
            self.regulons = pickle.load(open(regulons_fn, 'rb'))

    @staticmethod
    def read_file(fn):
        """
        Loading input files, supported file formats:
            * gef
            * gem
            * loom
            * h5ad
        Recommended formats:
            * h5ad
            * gef
        :param fn:
        :return:

        Example:
            grn.read_file('test.gef', bin_type='bins')
            or
            grn.read_file('test.h5ad')
        """
        extension = os.path.splitext(fn)[1]
        if extension == '.csv':
            raise TypeError('this method does not support csv files, '
                            'please read this file using functions outside of the InferenceRegulatoryNetwork class, '
                            'e.g. pandas.read_csv')
        elif extension == '.loom':
            data = sc.read_loom(fn)
            return data
        elif extension == '.h5ad':
            data = sc.read_h5ad(fn)
            return data

    def load_anndata_by_cluster(self, fn: str,
                                cluster_label: str,
                                target_clusters: list) -> an.AnnData:
        """
        When loading anndata, only load in wanted clusters
        One must perform Clustering beforehand
        :param fn: data file name
        :param cluster_label: where the clustering results are stored
        :param target_clusters: a list of interested cluster names
        :return:

        Example:
            sub_data = load_anndata_by_cluster(data, 'psuedo_class', ['HBGLU9'])
        """
        data = self.read_file(fn)
        if isinstance(data, an.AnnData):
            return data[data.obs[cluster_label].isin(target_clusters)]
        else:
            raise TypeError('data must be anndata.Anndata object')

    @staticmethod
    def is_valid_exp_matrix(mtx: pd.DataFrame):
        """
        check if the exp matrix is valid for the grn pipeline
        :param mtx:
        :return:
        """
        return (all(isinstance(idx, str) for idx in mtx.index)
                and all(isinstance(idx, str) for idx in mtx.columns)
                and (mtx.index.nlevels == 1)
                and (mtx.columns.nlevels == 1))

    @staticmethod
    def preprocess(adata: an.AnnData, min_genes=0, min_cells=3, min_counts=1, max_gene_num=4000):
        """
        Perform cleaning and quality control on the imported data before constructing gene regulatory network
        :param min_genes:
        :param min_cells:
        :param min_counts:
        :param max_gene_num:
        :return: a anndata.AnnData
        """
        adata.var_names_make_unique()  # compute the number of genes per cell (computes ‘n_genes' column)
        # # find mito genes
        sc.pp.ﬁlter_cells(adata, min_genes=0)
        # add the total counts per cell as observations-annotation to adata
        adata.obs['n_counts'] = np.ravel(adata.X.sum(axis=1))

        # ﬁltering with basic thresholds for genes and cells
        sc.pp.ﬁlter_cells(adata, min_genes=min_genes)
        sc.pp.ﬁlter_genes(adata, min_cells=min_cells)
        sc.pp.ﬁlter_genes(adata, min_counts=min_counts)
        adata = adata[adata.obs['n_genes'] < max_gene_num, :]
        return adata

    def uniq_genes(self, adjacencies):
        """
        Detect unique genes
        :param adjacencies:
        :return:
        """
        df = self._data.to_df()
        unique_adj_genes = set(adjacencies["TF"]).union(set(adjacencies["target"])) - set(df.columns)
        return unique_adj_genes

    @staticmethod
    def get_regulon_dict(regulon_list: Sequence[Regulon]) -> dict:
        """
        Form dictionary of { TF : Target } pairs from Regulons.
        :param regulon_list:
        :return:
        """
        assert regulon_list is not None, "regulons is not available, calculate regulons or load regulons results first"
        regulon_dict = {}
        for reg in regulon_list:
            targets = [target for target in reg.gene2weight]
            regulon_dict[reg.name] = targets
        return regulon_dict

    # Save to files
    def regulons_to_json(self, fn='regulons.json'):
        """
        Write regulon dictionary into json file
        :param fn:
        :return:
        """
        if not self.regulon_dict:
            self.regulon_dict = self.get_regulon_dict(self.regulons)
            self.data.uns['regulon_dict'] = self.regulon_dict
        with open(fn, 'w') as f:
            json.dump(self.regulon_dict, f, sort_keys=True, indent=4)

    # Regulons and Cell Types
    def cal_regulon_score(self, cluster_label='annotation', save_tmp=False, fn='regulon_specificity_scores.txt'):
        """
        Regulon specificity scores (RSS) across predicted cell types
        :param fn:
        :param save_tmp:
        :param cluster_label:
        :return:
        """
        rss_cellType = regulon_specificity_scores(self.auc_mtx, self.data.obs[cluster_label])
        if save_tmp:
            rss_cellType.to_csv(fn)
        self.rss = rss_cellType
        self.data.uns['rss'] = rss_cellType  # for each cell type
        return rss_cellType

    def get_top_regulons(self, cluster_label: str, topn: int) -> dict:
        """
        get top n regulons for each cell type based on regulon specificity scores (rss)
        :param cluster_label:
        :param topn:
        :return: a list
        """
        # Select the top 5 regulon_list from each cell type
        cats = sorted(list(set(self.data.obs[cluster_label])))
        topreg = {}
        for i, c in enumerate(cats):
            topreg[c] = list(self.rss.T[c].sort_values(ascending=False)[:topn].index)
        return topreg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date: Created on 21 Mar 2023 15:41
# @Author: Yao LI

import logging


class GetLogger:
    def __init__(self, path, clevel = logging.DEBUG, Flevel = logging.DEBUG):
        self.logger = logging.getLogger('spaGRN')
        self.logger.setLevel(logging.DEBUG)
        fmt = logging.Formatter(
            "[%(asctime)s][%(name)s][%(process)d][%(thread)d][%(module)s][%(lineno)d][%(levelname)s]: %(message)s",
            '%Y-%m-%d %H:%M:%S')
        # set CMD logging
        sh = logging.StreamHandler()
        sh.setFormatter(fmt)
        sh.setLevel(clevel)
        # set file logging
        fh = logging.FileHandler(path)
        fh.setFormatter(fmt)
        fh.setLevel(Flevel)
        self.logger.addHandler(sh)
        self.logger.addHandler(fh)

    def debug(self, message):
        self.logger.debug(message)

    def info(self, message):
        self.logger.info(message)

    def warning(self, message):
        self.logger.warning(message)

    def error(self, message):
        self.logger.error(message)

    def cri(self, message):
        self.logger.critical(message)


# logger = GetLogger('spaGRN.log', logging.ERROR, logging.DEBUG)
